import 'dart:convert';
import 'dart:io';

import 'package:puma/src/package_details.dart';
import 'package:puma/src/pubspec.dart';
import 'package:yaml/yaml.dart' as yaml;
import 'package:http/http.dart' as http;
import 'package:pub_semver/pub_semver.dart' as semver;

class Puma {
  final String hostUrl;
  final Pubspec pubspec;

  const Puma(this.hostUrl, this.pubspec);

  Future<Pubspec> install(String package, {bool dev = false}) async {
    final packageDetails = await _getPackageDetails(package);
    if (!dev) {
      return pubspec.addDependencies(packageDetails);
    }
    return pubspec.addDevDependencies(packageDetails);
  }

  Pubspec delete(String packageName) => pubspec
      .removeDependencies(packageName)
      .removeDevDependencies(packageName);

  Pubspec nextBreakingVersion() => pubspec.nextBreakingVersion();

  Pubspec nextMajorVersion() => pubspec.nextMajorVersion();

  Pubspec nextMinorVersion() => pubspec.nextMinorVersion();

  Pubspec nextPatchVersion() => pubspec.nextPatchVersion();

  Future<Pubspec> addFlutterAsset(String asset) async {
    final isDirectory = await FileSystemEntity.isDirectory(asset);
    if (isDirectory) {
      final d = Directory(asset);
      final paths = d.listSync(recursive: true)
        ..removeWhere((e) => e.statSync().type != FileSystemEntityType.file);
      final newAssets = paths.map((entity) => entity.path).toList();
      return pubspec.addFlutterAssets(newAssets);
    }
    return pubspec.addFlutterAsset(asset);
  }

  Future<PackageDetails> _getPackageDetails(String installInformation) async {
    if (installInformation.contains(':')) {
      final colonIndex = installInformation.indexOf(':');
      final packageName = installInformation.substring(0, colonIndex);
      final v = installInformation.substring(colonIndex + 1);
      if (v.startsWith('git')) {
        //  TODO(kevin): load git resource
        throw 'git is not currently supported';
      } else if (v.isNotEmpty && v != 'latest') {
        return _hostedPackage(packageName,
            version: semver.VersionConstraint.parse(v));
      } else {
        return _hostedPackage(packageName);
      }
    } else {
      return _localOrHostedPackage(installInformation);
    }
  }

  Future<PackageDetails> _localOrHostedPackage(
      String installInformation) async {
    final pathPackage = await _localPackage(installInformation);
    if (pathPackage != null) {
      return pathPackage;
    }
    return _hostedPackage(installInformation);
  }

  Future<PackageDetails> _localPackage(String packagePath) async {
    final normalizePath = packagePath.endsWith(Platform.pathSeparator)
        ? packagePath
        : '$packagePath${Platform.pathSeparator}';
    final pubspecPath = '${normalizePath}pubspec.yaml';
    final otherPubspec = File(pubspecPath);
    final pubspecExist = await otherPubspec.exists();
    if (pubspecExist) {
      final otherYaml =
          yaml.loadYaml(await otherPubspec.readAsString()) as yaml.YamlMap;
      return PathPackage(
        otherYaml['name'],
        normalizePath,
      );
    }
    return null;
  }

  Future<PackageDetails> _hostedPackage(String packageName,
      {semver.VersionConstraint version}) async {
    if (version == null) {
      return HostedPackage(
        packageName,
        await _getLatestStableVersion(packageName),
      );
    }
    if (await _checkVersionExist(packageName, version)) {
      return HostedPackage(
        packageName,
        version.toString(),
      );
    } else {
      //  TODO(kevin): add better error handling
      throw 'Version ${version.toString()} of package $packageName doesn\'t exist';
    }
  }

  Future<String> _getLatestStableVersion(String packageName) async {
    final response = await http.get('$hostUrl/api/documentation/$packageName');
    final packageInfos = json.decode(response.body);
    return '^${packageInfos['latestStableVersion']}';
  }

  Future<bool> _checkVersionExist(
      String packageName, semver.VersionConstraint version) async {
    final response = await http.get('$hostUrl/api/documentation/$packageName');
    final packageInfos = json.decode(response.body);
    return (packageInfos['versions'] as List).cast<Map>().firstWhere(
            (d) => version.allows(semver.Version.parse(d['version'])),
            orElse: () => null) !=
        null;
  }
}
